/* nm-openswan-service - openswan integration with NetworkManager
 *
 * Steve Hillier <keyhman@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * (C) Copyright 2007 Steve Hillier
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <glib/gi18n.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include <NetworkManager/NetworkManager.h>
#include <NetworkManager/NetworkManagerVPN.h>

//#include "openswan.h"
//#include "openswan/ipsec_policy.h"
//#include "constants.h"
/* only if we decide to compile whack.o and whacklib.o and link them to call direct 
#include "whack.h"
* for now we're gonna call out to /usr/libexec/ipsec/whack and see if that suffices 
* it will remove the version dependancy (in large part) from openswan and nm-openswan
*/
#include "nm-openswan-service.h"
#include "nm-utils.h"

typedef struct _ipsecProposal
{
	gchar * enc;
	gchar * auth;
} ipsecProposal;

typedef struct _Phase1data
{
	guint mode;
	gboolean dh1;
	gboolean dh2;
	gboolean dh5;
	guint32 keylife;
	gchar * LocalID;
	ipsecProposal *prop;
	GSList * propList;                       //for future multiple connections usage
	gint32   propList_Len;
} Phase1data;

typedef struct _Phase2data
{
	guint dhgroup;
	guint32 keylife;
	ipsecProposal *prop;
	GSList * propList;                        //for future use
	gint32   propList_Len;
} Phase2data;

typedef struct _ipsec_interface
{
	gchar * name;
	gboolean dhcpconfig;
	
	/* insert vars for manual iface config */
    gchar *         ip4_address;
    gchar *         ip4_netmask;
    gchar *         ip4_dns;
    gchar *         ip4_wins;
} ipsec_interface;


typedef struct _VPNConnData
{
    gchar          *name;
	gchar          *remote_Gateway;
	gchar          *remote_Network;
	gchar          *remote_SubnetMask;
	gchar          *auth_method;  /* PSK or CERT */
	gchar          *auth_secret;  /* pointer to shared secret or cert data */
	NMVPNState      ConnState;    /* For future use when nm can support multiple 
	                                 conncurrent connections */
	gboolean        XAuth;
	gboolean        promptXauth;
	gchar          *XAuth_username;
	gchar          *XAuth_password;
	
	gboolean        InterfaceMode;
	ipsec_interface *Interface;
	
	gboolean        AutoKeepAlive;
	guint32         KeepAliveFreq;
	gboolean        PFS;
	gboolean        DPD;
	gboolean        NAT_T;
	gboolean        ReplayDetection;
	
	Phase1data    *p1Data;
	gboolean       p1_success;
	Phase2data    *p2Data;
	gboolean       p2_success;
	
} VPNConnData;

typedef struct _NmOpenSwanData
{
	GMainLoop 		   *loop;
	DBusConnection 	   *con;
//	GIOChannel         *c_plutoSock;
	GIOChannel         *c_stdin;           /* in case the user wants to be prompted */
	GIOChannel         *c_stdout;          /* for X-Auth Username and Password */
	GIOChannel         *c_stderr;
	gchar              *plutoctl;          /* path to pluto socket */
	gchar              *whackbin;          /* path to openswan whack binary */
	NMVPNState		    state;
	GError              ConnError;
	GPid			    pid;
	guint			    quit_timer;
	guint			    helper_timer;
	GSList             *ConnList;            /* For naming conflict checks and future use */
	guint               ConnList_Len;
	VPNConnData        *ConnData;
	gboolean            nameConflict;
} NmOpenSwanData;


/*
 * temporary prompt_xauth helper function
 * prompts user for input (username and password)
 * so we can pass it to pluto
 */

gboolean prompt_xauth (gchar *username, gchar *password)
{
	gboolean HaveCredentials = FALSE;
	return HaveCredentials;
}


/*
 * args2buf helper function
 * converts GPtrArray into gchar *msgbuf[msg_args->len +2]
 * so we can pass it to whack (argv)
 */

static gchar* args2buf (gint len, GPtrArray *msg_args)
{
	if (len < 1)
		return NULL;
	
	guint i;
	gchar *msgbuf[len];
	
	for (i = 0; g_ptr_array_index(msg_args, i) ; i++)
	{
		msgbuf[i] = &(g_ptr_array_index(msg_args, i));
	}

	return &msgbuf;
}

//watch_io callback
/*
 * Watch stdout and stderr for messages from pluto 
 * while starting an ipsec connection
 *
 * We want to catch specific strings:
 * 1. update ConnState
 * 2. Capture error codes, or X-Auth codes
 */


static GIOFunc watch_io (GIOChannel *src, GIOCondition *cond, gpointer udata)
{
	NmOpenSwanData *data = (NmOpenSwanData *) udata;
	GError error;
	GIOStatus st;
	GIOError *err;
	gint retval;
	GString  *wckmsg;
	GString  *retcode;
	gboolean ConnUP = FALSE;

	wckmsg = g_string_new ("");
	retcode = g_string_new ("");
	
	//switch on IO signal (in or out)
	
	if (((GIOCondition*)cond == G_IO_IN) || ((GIOCondition*)cond == G_IO_PRI))
	{
		//st = g_io_channel_write_chars();
	}
	
	else if ((GIOCondition*)cond == G_IO_OUT)
	{
		st = g_io_channel_read_line_string (src, wckmsg, NULL, &error);
		//error checking on st 
		
		//we should preserve complete string for future use
		retcode = g_string_new (wckmsg->str);
		g_string_truncate (retcode, 3);   //we want the 3-number code
		retval = atoi (retcode->str);
		//parse retval for specific return codes we need to respond to (ie: xauth)
		//g_free (st);
		switch (retval)
		{
			case 002:
				//information about negotiation
				st = g_io_channel_write_chars (data->c_stdout, wckmsg->str, sizeof(wckmsg->str), NULL, &error);
				break;
			
			case 041:                 /* xauth username prompt */
				if (!data->ConnData->promptXauth)
				{
					st = g_io_channel_write_chars ( data->c_stdin, 
									   data->ConnData->XAuth_username,
									   sizeof(data->ConnData->XAuth_username),
									   NULL,
									   &error);
				}
				else
				{
					
					gint t;           //retries
					for (t = 3; t > 0; t--)
					{
						if (!prompt_xauth (&(data->ConnData->XAuth_username), &(data->ConnData->XAuth_password)))
							continue;	//failed to get credentials, try again
						else
						{
							if ( (strcmp (data->ConnData->XAuth_username, NULL) == 0) || (strcmp (data->ConnData->XAuth_password, NULL) == 0) )
								continue; 	// no empty passwords or usernames allowed in xauth, try again
							else
								break;
						}
					}

					st = g_io_channel_write_chars ( data->c_stdin, 
									   data->ConnData->XAuth_username,
									   sizeof(data->ConnData->XAuth_username),
									   NULL,
									   &error);
				}
				break;

			case 040:                 /* xauth password prompt */
				st = g_io_channel_write_chars ( data->c_stdin, 
								   data->ConnData->XAuth_password,
								   sizeof(data->ConnData->XAuth_password),
								   NULL,
								   &error);
				break;
			
			//catch case where XAuth has failed and report it!, then set state to down.

			case 004:
				//phase 1 and 2 successful return codes
				if (data->ConnData->p1_success)
				{
					data->ConnData->p2_success = TRUE;
					ConnUP = TRUE;
				}
				else
				{
					//phase 1 complete
					data->ConnData->p1_success = TRUE;
				}
				break;
				
			default:
				break;
		}
	}
	else if ((GIOCondition*)cond == G_IO_ERR)
	{
		st = g_io_channel_read_line_string (src, wckmsg, NULL, &error);
		nm_warning("Whack IO Error: %s", wckmsg->str);
	}
	else
		//unhandled GIOCondition
		nm_warning ("Warning: Unhandled GIOCondition in function watch_io");
	
	return ConnUP;
}




/* helper function to send the message out to whackbin
 *
 */


static gboolean send_message(gchar *msg[], ssize_t msglen, NmOpenSwanData *data)
{
    /* send to pluto */
	GError *error;
	GPid pid;
	gint stdin, stdout, stderr;
	
	/* this is where we spawn whack and attach to the IO for the callout */
	if (! (g_spawn_async_with_pipes( NULL,        /* workdir or should we make it "~" */
                              msg, /* argv */
                              NULL,        /* envp */
                              G_SPAWN_LEAVE_DESCRIPTORS_OPEN,        /* flags */
                              NULL,        /* child setup func */
                              NULL,        /* --- udata */
                              &pid,
                              &stdin, &stdout, &stderr,
                              error )) )      /* error */
	{
		nm_warning("CRITICAL -> Error spawning whack process: %s", error->message);
		return FALSE;
	}
	
	//assign callback for In/Out events on the channel
	data->c_stdin = g_io_channel_unix_new (stdin);
	data->c_stdout = g_io_channel_unix_new (stdout);
	data->c_stderr = g_io_channel_unix_new (stderr);
	
	g_io_add_watch (data->c_stdin, G_IO_IN, (GIOFunc) watch_io, (gpointer) data);
	g_io_add_watch (data->c_stdout, G_IO_OUT, (GIOFunc) watch_io, (gpointer) data);
	g_io_add_watch (data->c_stderr, G_IO_OUT, (GIOFunc) watch_io, (gpointer) data);	
	
	return TRUE;
}



/*
 * nm_openswan_dbus_signal_state_change
 *
 * Signal the bus that our state changed.
 *
 */
static void nm_openswan_dbus_signal_state_change (NmOpenSwanData *data, NMVPNState old_state)
{
	DBusMessage	*message;

	g_return_if_fail (data != NULL);

	if (!(message = dbus_message_new_signal (NM_DBUS_PATH_OPENSWAN, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_SIGNAL_STATE_CHANGE)))
	{
		nm_warning ("nm_openswan_dbus_signal_state_change(): Not enough memory for new dbus message!");
		return;
	}

	dbus_message_append_args (message, DBUS_TYPE_UINT32, &old_state, DBUS_TYPE_UINT32, &(data->state), DBUS_TYPE_INVALID);

	if (!dbus_connection_send (data->con, message, NULL))
		nm_warning ("nm_openswan_dbus_signal_state_change(): Could not raise the signal!");

	dbus_message_unref (message);
}



/*
 * nm_openswan_set_state
 *
 * Helper Function to Set our state and make sure to signal the bus.
 *
 */
static void nm_openswan_set_state (NmOpenSwanData *data, NMVPNState new_state)
{
	NMVPNState	old_state;

	g_return_if_fail (data != NULL);

	old_state = data->state;

	if (old_state != new_state)
	{
		data->state = new_state;
		nm_openswan_dbus_signal_state_change (data, old_state);
	}
}



/*
 * handle_stop_vpn
 *
 */

static gboolean nm_openswan_dbus_handle_stop_vpn (NmOpenSwanData *data)
{
	GPtrArray *msg_args;
	gchar *conn_name;
	gchar *act_term;
	gchar *act_unroute;
	gchar *act_delete;
	gchar *msgbuf;
	
	conn_name = g_strdup_printf("--name ");
	act_term = g_strdup_printf(" --terminate");
	act_unroute = g_strdup_printf(" --unroute");
	act_delete = g_strdup_printf(" --delete");

	
	//contruct conn terminate message for whack
	msg_args = g_ptr_array_new();	
	g_ptr_array_add (msg_args, data->whackbin);
	g_ptr_array_add (msg_args, conn_name);
	g_ptr_array_add (msg_args, data->ConnData->name);
	g_ptr_array_add (msg_args, act_term);
	g_ptr_array_add(msg_args, "NULL");
	
	msgbuf = args2buf (msg_args->len, msg_args);
	nm_info("Here's my whack-terminate command string: \n\n/usr/libexec/ipsec/whack %s", &msgbuf);

	//send message to whack
	if (!send_message (&msgbuf, sizeof(msgbuf), data))
		return FALSE;
	
	/* construct conn unroute message 
	 * pluto sets up conns to opperunistically re-initiate by preserving 
	 * the routing table entry and caching the previous connection data
	 * to avoid unexpectedly re-intiating a connection, we send an unroute 
	 * command string to pluto via whack */
	
	g_free (msgbuf);
	g_ptr_array_free (msg_args, TRUE);
	
	msg_args = g_ptr_array_new();
	g_ptr_array_add (msg_args, data->whackbin);
	g_ptr_array_add (msg_args, conn_name);
	g_ptr_array_add (msg_args, data->ConnData->name);
	g_ptr_array_add (msg_args, act_unroute);
	g_ptr_array_add(msg_args, "NULL");
	
	msgbuf = args2buf (msg_args->len, msg_args);
	nm_info("Here's my whack-unroute command string: \n\n %s", &msgbuf);
	//send message to whack
	if (!send_message (&msgbuf, sizeof(msgbuf), data))
		return FALSE;

	
	/* construct conn delete message 
	 * pluto caches connection definitions it's fed into an internal
	 * connection Database. No two entries can have the same "name".
	 * to avoid possible naming conflicts, we delete the connection definition
	 * from pluto's internal Connection database */
	
	g_free (msgbuf);
	g_ptr_array_free (msg_args, TRUE);

	msg_args = g_ptr_array_new();
	g_ptr_array_add (msg_args, data->whackbin);
	g_ptr_array_add (msg_args, conn_name);
	g_ptr_array_add (msg_args, data->ConnData->name);
	g_ptr_array_add (msg_args, act_delete);
	g_ptr_array_add(msg_args, "NULL");
	
	msgbuf = args2buf (msg_args->len, msg_args);
	nm_info("Here's my whack-delete command string: \n\n %s", &msgbuf);
	//send message to whack
	if (!send_message (&msgbuf, sizeof(msgbuf), data))
		return FALSE;

	nm_openswan_set_state ( data, NM_VPN_STATE_STOPPED);
	return TRUE; 
}

/*
 * nm_dbus_create_error_message
 *
 * Make a DBus error message
 *
 */
static DBusMessage *nm_dbus_create_error_message (DBusMessage *message, const char *exception_namespace,
										const char *exception, const char *format, ...)
{
	gchar *exception_text;
	DBusMessage	*reply;
	va_list		 args;
	gchar			 error_text[512];

	va_start (args, format);
	vsnprintf (error_text, 512, format, args);
	va_end (args);

	exception_text = g_strdup_printf ("%s.%s", exception_namespace, exception);
	reply = dbus_message_new_error (message, exception_text, error_text);
	g_free (exception_text);

	return (reply);
}


/*
 * nm_openswan_dbus_signal_failure
 *
 * Signal the bus that some VPN operation failed.
 *
 */
static void nm_openswan_dbus_signal_failure (NmOpenSwanData *data, const char *signal)
{
	DBusMessage	*message;
	const gchar	*error_msg = NULL;

	g_return_if_fail (data != NULL);
	g_return_if_fail (signal != NULL);

	if (!strcmp (signal, NM_DBUS_VPN_SIGNAL_LOGIN_FAILED))
		error_msg = _("The VPN login failed because the user name and password were not accepted.");
	else if (!strcmp (signal, NM_DBUS_VPN_SIGNAL_LAUNCH_FAILED))
		error_msg = _("The VPN login failed because the VPN program could not be started.");
	else if (!strcmp (signal, NM_DBUS_VPN_SIGNAL_CONNECT_FAILED))
		error_msg = _("The VPN login failed because the VPN program could not connect to the VPN server.");
	else if (!strcmp (signal, NM_DBUS_VPN_SIGNAL_VPN_CONFIG_BAD))
		error_msg = _("The VPN login failed because the VPN configuration options were invalid.");
	else if (!strcmp (signal, NM_DBUS_VPN_SIGNAL_IP_CONFIG_BAD))
		error_msg = _("The VPN login failed because the VPN program received an invalid configuration from the VPN server.");

	if (!error_msg)
		return;

	if (!(message = dbus_message_new_signal (NM_DBUS_PATH_OPENSWAN, NM_DBUS_INTERFACE_OPENSWAN, signal)))
	{
		nm_warning ("Not enough memory for new dbus message!");
		return;
	}

	dbus_message_append_args (message, DBUS_TYPE_STRING, &error_msg, DBUS_TYPE_INVALID);
	if (!dbus_connection_send (data->con, message, NULL))
		nm_warning ("Could not raise the signal!");

	dbus_message_unref (message);
}




/*
 * nm_openswan_quit_timer_cb
 *
 * Callback to quit nm-openswan-service after a certain period of time.
 *
 */
static gboolean nm_openswan_quit_timer_cb (NmOpenSwanData *data)
{
	data->quit_timer = 0;

	g_return_val_if_fail (data != NULL, FALSE);

	g_main_loop_quit (data->loop);

	return FALSE;
}


/*
 * nm_openswan_schedule_quit_timer
 *
 * If openswan isn't running, and we haven't been asked to do anything in a while,
 * then we just exit since NetworkManager will re-launch us later.
 *
 */
static void nm_openswan_schedule_quit_timer (NmOpenSwanData *data, guint interval)
{
	g_return_if_fail (data != NULL);

	if (data->quit_timer == 0)
		data->quit_timer = g_timeout_add (interval, (GSourceFunc) nm_openswan_quit_timer_cb, data);
}


/*
 * nm_openswan_cancel_quit_timer
 *
 * Cancel a quit timer that we've scheduled before.
 *
 */
static void nm_openswan_cancel_quit_timer (NmOpenSwanData *data)
{
	g_return_if_fail (data != NULL);

	if (data->quit_timer > 0)
		g_source_remove (data->quit_timer);
}


/*
 * nm_openswan_helper_timer_cb
 *
 * If we haven't received the IP4 config info from the helper before the timeout
 * occurs, we kill openswan.
 *
 */
static gboolean nm_openswan_helper_timer_cb (NmOpenSwanData *data)
{
	data->helper_timer = 0;

	g_return_val_if_fail (data != NULL, FALSE);

	nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_CONNECT_FAILED);
	nm_openswan_dbus_handle_stop_vpn (data);

	return FALSE;
}


/*
 * nm_openswan_schedule_helper_timer
 *
 * Once openswan is running, we wait for the helper to return the IP4 configuration
 * information to us.  If we don't receive that information within 7 seconds,
 * we kill openswan.
 *
 */
static void nm_openswan_schedule_helper_timer (NmOpenSwanData *data)
{
	g_return_if_fail (data != NULL);

	if (data->helper_timer == 0)
		data->helper_timer = g_timeout_add (10000, (GSourceFunc) nm_openswan_helper_timer_cb, data);
}


/*
 * nm_openswan_cancel_helper_timer
 *
 * Cancel a helper timer that we've scheduled before.
 *
 */
static void nm_openswan_cancel_helper_timer (NmOpenSwanData *data)
{
	g_return_if_fail (data != NULL);

	if (data->helper_timer > 0)
		g_source_remove (data->helper_timer);
}


/*
 * pluto_watch_cb
 *
 * Watch our child openswan process and get notified of events from it.
 *
 */
static void pluto_watch_cb (GPid pid, gint status, gpointer user_data)
{
	guint	error = -1;

	NmOpenSwanData *data = (NmOpenSwanData *)user_data;

	if (WIFEXITED (status))
	{
		error = WEXITSTATUS (status);
		if (error != 0)
			nm_warning ("pluto exited with error code %d", error);
	}
	else if (WIFSTOPPED (status))
		nm_warning ("pluto stopped unexpectedly with signal %d", WSTOPSIG (status));
	else if (WIFSIGNALED (status))
		nm_warning ("pluto died with signal %d", WTERMSIG (status));
	else
		nm_warning ("pluto died from an unknown cause");

	/* Reap child if needed. */
	waitpid (data->pid, NULL, WNOHANG);
	data->pid = 0;

	/* Must be after data->state is set since signals use data->state */
	switch (error)
	{
		case 2:	/* Couldn't log in due to bad user/pass */
			nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_LOGIN_FAILED);
			break;

		case 1:	/* Other error (couldn't bind to address, etc) */
			nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_CONNECT_FAILED);
			break;

		default:
			break;
	}

	nm_openswan_set_state (data, NM_VPN_STATE_STOPPED);
	nm_openswan_schedule_quit_timer (data, 10000);
}


/*
 * conn_watch_cb
 *
 * Watch our ipsec control process as it does an operation
 * and get notified of events from it.
 *
 */
static void conn_watch_cb (GPid pid, gint status, gpointer user_data)
{
	guint	error = -1;

	NmOpenSwanData *data = (NmOpenSwanData *) user_data;
	
	if (WIFEXITED (status))
	{
		error = WEXITSTATUS (status);
		if (error != 0)
			nm_warning ("ipsec control exited with error code %d", error);
	}
	else if (WIFSTOPPED (status))
		nm_warning ("ipsec control stopped unexpectedly with signal %d", WSTOPSIG (status));
	else if (WIFSIGNALED (status))
		nm_warning ("ipsec control died with signal %d", WTERMSIG (status));
	else
		nm_warning ("ipsec control died from an unknown cause");

	/* Reap child if needed. */
	waitpid (data->pid, NULL, WNOHANG);
	data->pid = 0;

	/* Must be after data->state is set since signals use data->state */
	switch (error)
	{
		case 2:	/* Couldn't log in due to bad user/pass */
			nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_LOGIN_FAILED);
			break;

		case 1:	/* Other error (couldn't bind to address, etc) */
			nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_CONNECT_FAILED);
			break;

		default:
			break;
	}

	nm_openswan_set_state (data, NM_VPN_STATE_STOPPED);
	nm_openswan_schedule_quit_timer (data, 10000);
}


/*
 * init connection to pluto service process
 * THIS FUNCTION NEEDS TO BE COMPLETELY RE-WRITTEN
 */
static gboolean * nm_openswan_init_IPC (NmOpenSwanData *data)
{
	gint stdin, stdout, stderr;
	GPid pid;
	
	//check to see if pluto is running
	/* Find pluto control socket */
	data->plutoctl = plutoctl_paths;
	while (*(data->plutoctl) != NULL)
	{
		if (g_file_test (*(data->plutoctl), G_FILE_TEST_EXISTS))
	 	 	break;
	 	(data->plutoctl)++;
	}

	if (!data->plutoctl)
	{
	 	nm_info ("Could not find Pluto IPC Socket! Is OpenSwan running?");
	 	return FALSE;
	}

	/* Find the whack binary to call out */
	data->whackbin = whackbin_paths;
	while (*(data->whackbin) != NULL)
	{
		if (g_file_test (*(data->whackbin), G_FILE_TEST_EXISTS))
	 	 	break;
	 	(data->whackbin)++;
	}

	if (!data->whackbin)
	{
	 	nm_info ("Could not find Whack Binary!");
	 	return FALSE;
	}
	
	//okay, we have what we need to attempt a connection
	return TRUE;
}


/*
 *
 *  Get Current Connections Database from running pluto process 
 * This way we can check for name conflicts on the incoming connection configs
 *
 */

static guint pluto_get_ConnList (NmOpenSwanData *data)
{
	if (!data->ConnList)
	{
		return 0;
	}
	else
	{
		//create message
		//pack messages
		//send message to pluto
		return g_slist_length(data->ConnList);
	}
}



/*
 * nm_openswan_start_vpn_connection
 *
 * Start the IPsec connection specified by data->ConnName.
 *
 */
static gboolean nm_openswan_start_vpn_connection (NmOpenSwanData *data)
{
	GPtrArray *msg_args;
	gint      msg_len = -1;
	GError  *error = NULL;
	
	gchar *conn_name = "--name";
	gchar *act_add = "--add";
	gchar *act_init = "--initiate";

	g_return_val_if_fail (data != NULL, FALSE);
	
	/* Whack Message Format:
ipsec whack --name novalab --encrypt --tunnel --ike "aes128-sha1-96" --esp "aes128-sha1-96" --pfs --dpddelay "10" --dpdtimeout "120" --dpdaction "clear" --ikelifetime "28800" --psk \
        --host "172.22.22.50"   --nexthop "%direct" --updown "ipsec _updown"    --sendcert "always"       \
        --to --host "209.217.101.155" --client "10.1.1.0/24"  --nexthop "%direct" --updown "ipsec _updown"    --sendcert "always"       \
        --ipseclifetime "1800" --rekeymargin "540" \
        --keyingtries "0"    \
	*/
	
	//LEFT SIDE (ME)
	gchar *flag_name = " --name ";
	//add value to string
	gchar *flag_encrypt = " --encrypt ";
	gchar *flag_authenticate = " --authenticate ";
	gchar *flag_tunnel = " --tunnel ";
	gchar *flag_psk = " --psk ";
	gchar *flag_rsasig = " --rsasig ";
	gchar *flag_ike = " --ike ";
	//add p1pol
	gchar *flag_esp = " --esp ";
	//add p2pol
	gchar *flag_pfs = " --pfs ";
	gchar *flags_dpd = " --dpddelay \"10\" --dpdtimeout \"120\" ";
	gchar *flag_myip = " --host ";
	//add my ip, or %any identifier
	gchar *flag_nexthop = " --nexthop \"\%direct\" ";
	gchar *flag_updown = " --updown ipsec _updown ";
	gchar *flag_sendcert = " --sendcert \"always\" ";
	//RIGHT SIDE (Target)
	gchar *flag_to = " -to ";
	gchar *flag_remote_gateway = " --host ";
	//add remote gateway
	gchar *flag_remote_subnet = " --client ";
	//add remote network  / netmask
	gchar *flag_ikelifetime = " --ikelifetime ";
	gchar *flag_ipseclifetime = " --ipseclifetime ";
	gchar *flag_rekeymargin = " --rekeymargin \"540\" ";
	gchar *flag_keyingtries = " --keyingtries \"0\" ";
	
	//process Connection Data
	gchar *msgbuf;
	gint i;
	msg_args = g_ptr_array_new();
	
	g_ptr_array_add(msg_args, data->whackbin);
	
	g_ptr_array_add(msg_args, flag_name);
	g_ptr_array_add(msg_args, data->ConnData->name);
	g_ptr_array_add(msg_args, act_add); //the action Whack will take -- add conn def to conn DB
	g_ptr_array_add(msg_args, flag_encrypt);
	g_ptr_array_add(msg_args, flag_authenticate);
	g_ptr_array_add(msg_args, flag_tunnel);
	if ((strcmp(&data->ConnData->auth_method, "PSK")) == 0) 
	{
		g_ptr_array_add(msg_args, flag_psk);
		g_ptr_array_add(msg_args, data->ConnData->auth_secret);
	}
	else if ((strcmp(&data->ConnData->auth_method, "CERT")) == 0) 
	{
		g_ptr_array_add(msg_args, flag_rsasig);
		g_ptr_array_add(msg_args, data->ConnData->auth_secret);
	}
	else
		return FALSE;
	
	g_ptr_array_add(msg_args, flag_ike);
	g_ptr_array_add(msg_args, data->ConnData->p1Data->prop);
	g_ptr_array_add(msg_args, flag_esp);
	g_ptr_array_add(msg_args, data->ConnData->p2Data->prop);
	g_ptr_array_add(msg_args, flag_sendcert);
	g_ptr_array_add(msg_args, flag_nexthop);
	g_ptr_array_add(msg_args, flag_updown);
	
	if (data->ConnData->PFS)
		g_ptr_array_add(msg_args, flag_pfs);
	if (data->ConnData->DPD)
		g_ptr_array_add(msg_args, flags_dpd);
	
	//start the Right Side
	g_ptr_array_add(msg_args, flag_remote_gateway);
	g_ptr_array_add(msg_args, data->ConnData->remote_Gateway);
	g_ptr_array_add(msg_args, flag_remote_subnet);
	g_ptr_array_add(msg_args, data->ConnData->remote_Network);
	g_ptr_array_add(msg_args, "/");
	g_ptr_array_add(msg_args, data->ConnData->remote_SubnetMask);
	g_ptr_array_add(msg_args, flag_sendcert);
	g_ptr_array_add(msg_args, flag_nexthop);
	g_ptr_array_add(msg_args, flag_updown);
	//add null terminator
	g_ptr_array_add(msg_args, "NULL");
	/* ok, to call send_message we need to create a null terminated array of strings */

	msgbuf = args2buf (msg_args->len, msg_args);
	nm_info("Here's my whack-add command string: \n\n %s", &msgbuf);
	if (!send_message (&msgbuf, sizeof(msgbuf), data))
		return FALSE;

	
	g_free(msgbuf);
	g_ptr_array_free(msg_args, TRUE);
	
	msg_args = g_ptr_array_new();
	g_ptr_array_add (msg_args, data->whackbin);
	g_ptr_array_add (msg_args, conn_name);
	g_ptr_array_add (msg_args, data->ConnData->name);
	g_ptr_array_add (msg_args, act_init);
	g_ptr_array_add (msg_args, "NULL");
	
	msgbuf = args2buf (msg_args->len, msg_args);
	nm_info("Here's my whack-init command string: \n\n %s", &msgbuf);
	if (!send_message (&msgbuf, sizeof(msgbuf), data))
		return FALSE;
	
	nm_openswan_set_state (data, NM_VPN_STATE_STARTED);
	return TRUE; 
}


typedef enum OptType
{
	OPT_TYPE_UNKNOWN = 0,
	OPT_TYPE_ADDRESS,
	OPT_TYPE_INT,
	OPT_TYPE_ASCII,
	OPT_TYPE_BOOLEAN,
	OPT_TYPE_NONE
} OptType;

typedef struct Option
{
	const char *name;
	OptType type;
} Option;

/*
 * nm_openswan_config_options_validate
 *
 * Make sure the config options are sane
 *
 */
static gboolean nm_openswan_config_options_validate (char **data_items, int num_items, NmOpenSwanData *data)
{
	Option	allowed_opts[] = {
		{ NMOS_STRING_NAME,		              OPT_TYPE_ASCII },
		{ NMOS_STRING_REMOTE_GATEWAY,         OPT_TYPE_ADDRESS },               /* Remote gateway */
		{ NMOS_STRING_REMOTE_NETWORK,		  OPT_TYPE_ADDRESS },               /* Remote Network */
		{ NMOS_STRING_REMOTE_SUBNET_MASK,     OPT_TYPE_ADDRESS },               /* Subnet Mask */
		{ NMOS_STRING_AUTH_METHOD,		      OPT_TYPE_ASCII },                 /* Auth Method */
	    { NMOS_STRING_AUTH_SECRET,            OPT_TYPE_ASCII },
		{ NMOS_STRING_XAUTH,		          OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_XAUTH_USERNAME,		  OPT_TYPE_ASCII },
		{ NMOS_STRING_XAUTH_PASSWORD,	 	  OPT_TYPE_ASCII },
		{ NMOS_STRING_PFS,		              OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_DPD,		              OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_NATT,		              OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_REPLAY_DETECTION,	      OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_AUTO_KEEPALIVE,		  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_KEEP_ALIVE_FREQUENCY,   OPT_TYPE_INT },
		{ NMOS_STRING_INTERFACE_MODE,		  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_INTERFACE_NAME,		  OPT_TYPE_ASCII },
		{ NMOS_STRING_INTERFACE_DHCP,		  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_INTERFACE_IP_ADDRESS,   OPT_TYPE_ADDRESS },
		{ NMOS_STRING_INTERFACE_SUBNET_MASK,  OPT_TYPE_ADDRESS },
		{ NMOS_STRING_INTERFACE_DNS_SERVER,   OPT_TYPE_ADDRESS },
		{ NMOS_STRING_PHASE1_MODE,		      OPT_TYPE_ADDRESS },
		{ NMOS_STRING_PHASE1_DH_GROUP_1,	  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_PHASE1_DH_GROUP_2,	  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_PHASE1_DH_GROUP_5,	  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_PHASE1_KEYLIFE,		  OPT_TYPE_BOOLEAN },
		{ NMOS_STRING_PHASE1_LOCAL_ID,		  OPT_TYPE_INT },
		{ NMOS_STRING_PHASE1_AUTH,		      OPT_TYPE_ASCII },
		{ NMOS_STRING_PHASE1_ENC,		      OPT_TYPE_ASCII },
		{ NMOS_STRING_PHASE2_DH_GROUP,		  OPT_TYPE_ASCII },
		{ NMOS_STRING_PHASE2_KEYLIFE,		  OPT_TYPE_INT },
		{ NMOS_STRING_PHASE2_AUTH,		      OPT_TYPE_INT },
		{ NMOS_STRING_PHASE2_ENC,             OPT_TYPE_ASCII },
		{ NULL,                       		  OPT_TYPE_UNKNOWN }
	};
	guint	 i;
	gint     v;
	gboolean b;

	g_return_val_if_fail (data_items != NULL, FALSE);
	g_return_val_if_fail (num_items >= 2, FALSE);

	for (i = 0; i < num_items; i += 2)
	{
		Option *opt = NULL;
		unsigned int t, len;
		char *opt_value;

		if (!data_items[i] || !data_items[i+1])
			return FALSE;
		opt_value = data_items[i+1];

		/* Find the option in the allowed list */
		for (t = 0; t < sizeof (allowed_opts) / sizeof (Option); t++)
		{
			opt = &allowed_opts[t];
			if (opt->name && !strcmp (opt->name, data_items[i]))
				break;
		}
		if (!opt->name)	/* not found */
		{
			nm_warning ("VPN option '%s' is not allowed.", data_items[i]);
			return FALSE;
		}

		/* Don't allow control characters at all */
		len = strlen (opt_value);
		for (t = 0; t < len; t++)
		{
			if (iscntrl (opt_value[t]))
			{
				nm_warning ("There were invalid characters in the VPN option '%s' - '%s'.", data_items[i], opt_value);
				return FALSE;
			}
		}

		/* Validate the option's data */
		switch (opt->type)
		{
			case OPT_TYPE_ASCII:
				/* What other characters should we reject?? */
				/* Process String Args */
				if ((strcmp(opt->name, NMOS_STRING_NAME)) == 0)
						data->ConnData->name = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_AUTH_METHOD)) == 0)
						data->ConnData->auth_method = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_AUTH_SECRET)) == 0)
						data->ConnData->auth_secret = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_XAUTH_USERNAME)) == 0)
						data->ConnData->XAuth_username = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_XAUTH_PASSWORD)) == 0)
						data->ConnData->XAuth_password = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_INTERFACE_NAME)) == 0)
						data->ConnData->Interface->name = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_INTERFACE_DHCP)) == 0)
						data->ConnData->p1Data->mode = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_PHASE1_LOCAL_ID)) == 0)
						data->ConnData->p1Data->LocalID = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_PHASE1_AUTH)) == 0)
						data->ConnData->p1Data->prop->auth = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_PHASE1_ENC)) == 0)
						data->ConnData->p1Data->prop->enc = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_PHASE2_AUTH)) == 0)
						data->ConnData->p2Data->keylife = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_PHASE2_ENC)) == 0)
						data->ConnData->p2Data->keylife = opt_value;
				else
					return FALSE;
				
				break;

			case OPT_TYPE_INT:
				/* These have numerical data in string format */
				/* process Numerical Args */
				v = -1;
				if (! (v = atoi(opt_value)))
				{
					nm_warning("Numerical Options can only contain integer Data. I got %s for option %s.", &opt_value, &opt->name);
					return FALSE;
				}
				else 
				{
					//valid numerical arg, find it's data holder and store it
					if ((strcmp(opt->name, NMOS_STRING_KEEP_ALIVE_FREQUENCY)) == 0)
						data->ConnData->KeepAliveFreq = v;
					else if ((strcmp(opt->name, NMOS_STRING_PHASE1_KEYLIFE)) == 0)
						data->ConnData->p1Data->keylife = v;
					else if ((strcmp(opt->name, NMOS_STRING_PHASE2_DH_GROUP)) == 0)
						data->ConnData->p2Data->dhgroup = v;
					else if ((strcmp(opt->name, NMOS_STRING_PHASE2_KEYLIFE)) == 0)
						data->ConnData->p2Data->keylife = v;
					else
						return FALSE;
				}
				break;

			case OPT_TYPE_BOOLEAN:
				/* These have strings: TRUE or FALSE */
				/* Process Boolean Args */
				if ( (!strcmp (opt_value, "TRUE")) == 0)
					b = TRUE;
				else if ( (!strcmp (opt_value, "FALSE")) == 0)
					b = FALSE;
				else {
					nm_warning("Boolean Flags can only be TRUE or FALSE. I got %s for option %s.", &opt_value, &opt->name);
					return FALSE;
				}
				//flags look good, lets find a container for the data
				if ((strcmp(&opt->name, NMOS_STRING_XAUTH)) == 0)
					data->ConnData->XAuth = b;
				else if ((strcmp (opt->name, NMOS_STRING_PFS)) == 0)
					data->ConnData->PFS = b;
				else if ((strcmp (opt->name, NMOS_STRING_DPD)) == 0)
					data->ConnData->DPD = b;
				else if ((strcmp (opt->name, NMOS_STRING_NATT)) == 0)
					data->ConnData->NAT_T = b;
				else if ((strcmp (opt->name, NMOS_STRING_REPLAY_DETECTION)) == 0)
					data->ConnData->ReplayDetection = b;
				else if ((strcmp (opt->name, NMOS_STRING_AUTO_KEEPALIVE)) == 0)
					data->ConnData->AutoKeepAlive = b;
				else if ((strcmp (opt->name, NMOS_STRING_INTERFACE_MODE)) == 0)
					data->ConnData->InterfaceMode = b;
				else if ((strcmp (opt->name, NMOS_STRING_INTERFACE_DHCP)) == 0)
					data->ConnData->Interface->dhcpconfig = b;
				else if ((strcmp (opt->name, NMOS_STRING_PHASE1_DH_GROUP_1)) == 0)
					data->ConnData->p1Data->dh1 = b;
				else if ((strcmp (opt->name, NMOS_STRING_PHASE1_DH_GROUP_2)) == 0)
					data->ConnData->p1Data->dh2 = b;
				else if ((strcmp (opt->name, NMOS_STRING_PHASE1_DH_GROUP_5)) == 0)
					data->ConnData->p1Data->dh5 = b;
				else
					return FALSE;
				break;

			case OPT_TYPE_NONE:
				/* These have blank data */
				break;

			case OPT_TYPE_ADDRESS:
				/* Can be any legal hostname or IP address */
				if ((strcmp(&opt->name, NMOS_STRING_REMOTE_GATEWAY)) == 0)
					data->ConnData->remote_Gateway = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_REMOTE_NETWORK)) == 0)
					data->ConnData->remote_Network = opt_value;
				else if ((strcmp(opt->name, NMOS_STRING_REMOTE_SUBNET_MASK)) == 0)
					data->ConnData->remote_SubnetMask = opt_value;
				if (data->ConnData->InterfaceMode)
				{
					if ((strcmp(opt->name, NMOS_STRING_INTERFACE_IP_ADDRESS)) == 0)
						data->ConnData->Interface->ip4_address = opt_value;
					else if ((strcmp(opt->name, NMOS_STRING_INTERFACE_SUBNET_MASK)) == 0)
						data->ConnData->Interface->ip4_netmask = opt_value;
					else if ((strcmp(opt->name, NMOS_STRING_INTERFACE_DNS_SERVER)) == 0)
						data->ConnData->Interface->ip4_dns = opt_value;
					else if ((strcmp(opt->name, NMOS_STRING_INTERFACE_WINS_SERVER)) == 0)
						data->ConnData->Interface->ip4_wins = opt_value;	
				}
				else
					return FALSE;
				
				break;

			default:
				return FALSE;
				break;
		}
	}
	return TRUE;
}



//check for conn name conflicts (each name must be unique)

static void name_conflict_check (NmOpenSwanData *data)
{
/*	
	VPNConnData * cdata = (VPNConnData *) g_slist_next(data->ConnList);
	
	if ( strcmp (data->ConnData->name, cdata->name) == 0)
		data->nameConflict = TRUE;
	else 
*/											  
	data->nameConflict = FALSE;
	return;
}


/*
 * nm_openswan_dbus_handle_start_vpn
 *
 * Parse message arguments and start the VPN connection.
 *
 */
static gboolean nm_openswan_dbus_handle_start_vpn (DBusMessage *message, NmOpenSwanData *data)
{
	gboolean success = FALSE;
	gchar **data_items = NULL;
	gint  num_items = -1;
	DBusError error;
	
	g_return_val_if_fail (message != NULL, FALSE);
	g_return_val_if_fail (data != NULL, FALSE);

	//set the state.
	nm_openswan_set_state (data, NM_VPN_STATE_STARTING);

	dbus_error_init (&error);
	if (!dbus_message_get_args (message, &error,
					DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &data_items, &num_items,
				    DBUS_TYPE_INVALID))
	{
		nm_warning ("Could not process the request because its arguments were invalid.  dbus said: '%s'", error.message);
		nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_VPN_CONFIG_BAD);
		dbus_error_free (&error);
		goto out;
	}

	if (!nm_openswan_config_options_validate (data_items, num_items, data))
	{
		nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_VPN_CONFIG_BAD);
		goto out;
	}
	
	/* insert check for connection name conflict
	 * If exists, then check state of connection
	 * and we'll need to replace the current definition with the parameters we 
	 * just recieved.
	
	if (data->ConnList_Len > 0)
	{
		g_slist_foreach (data->ConnList, (GFunc)name_conflict_check, data);
	    if (data->nameConflict)
	    {
			//we have a naming conflict - this should NEVER Happen, but just in case.
			nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_VPN_CONFIG_BAD);
			goto out;
		}
	}
*/	
	/* New ConnData is unique, Now we can finally try to activate the VPN */
	
	if ( nm_openswan_start_vpn_connection (data) )
	{
		success = TRUE;
		goto out;
	}
	else
	{
		nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_LAUNCH_FAILED);
		goto out;
	}

out:
	if (!success)
		nm_openswan_set_state (data, NM_VPN_STATE_STOPPED);
	else
		nm_openswan_set_state (data, NM_VPN_STATE_STARTED);
	
	return success;
}



/*
 * nm_openswan_dbus_start_vpn_connection
 *
 * Begin a VPN connection.
 *
 */
static DBusMessage *nm_openswan_dbus_start_vpn (DBusConnection *con, DBusMessage *message, NmOpenSwanData *data)
{
	DBusMessage		*reply = NULL;

	g_return_val_if_fail (data != NULL, NULL);
	g_return_val_if_fail (con != NULL, NULL);
	g_return_val_if_fail (message != NULL, NULL);

	switch (data->state)
	{
		case NM_VPN_STATE_STARTING:
			reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_STARTING_IN_PROGRESS,
						"Could not process the request because the VPN connection is already being started.");
			break;

		case NM_VPN_STATE_STARTED:
			reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_ALREADY_STARTED,
						"Could not process the request because a VPN connection was already active.");
			break;

		case NM_VPN_STATE_STOPPING:
			reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_STOPPING_IN_PROGRESS,
						"Could not process the request because the VPN connection is being stopped.");
			break;

		case NM_VPN_STATE_STOPPED:
		{
			nm_openswan_cancel_quit_timer (data);	
			nm_openswan_dbus_handle_start_vpn (message, data);
			reply = dbus_message_new_method_return (message);
			break;
		}
		default:
			g_assert_not_reached();
			break;
	}

	return reply;
}


/*
 * nm_openswan_dbus_stop_vpn
 *
 * Terminate a VPN connection.
 *
 */
static DBusMessage *nm_openswan_dbus_stop_vpn (DBusConnection *con, DBusMessage *message, NmOpenSwanData *data)
{
	DBusMessage		*reply = NULL;

	g_return_val_if_fail (data != NULL, NULL);
	g_return_val_if_fail (con != NULL, NULL);
	g_return_val_if_fail (message != NULL, NULL);

	switch (data->state)
	{
		case NM_VPN_STATE_STOPPING:
			reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_STOPPING_IN_PROGRESS,
						"Could not process the request because the VPN connection is already being stopped.");
			break;

		case NM_VPN_STATE_STOPPED:
			reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_ALREADY_STOPPED,
						"Could not process the request because no VPN connection was active.");
			break;

		case NM_VPN_STATE_STARTING:
		case NM_VPN_STATE_STARTED:
			nm_openswan_dbus_handle_stop_vpn (data);
			reply = dbus_message_new_method_return (message);
			break;

		default:
			g_assert_not_reached();
			break;
	}

	return reply;
}


/*
 * nm_openswan_dbus_get_state
 *
 * Return some state information to NetworkManager.
 *
 */
static DBusMessage *nm_openswan_dbus_get_state (DBusConnection *con, DBusMessage *message, NmOpenSwanData *data)
{
	DBusMessage		*reply = NULL;

	g_return_val_if_fail (data != NULL, NULL);
	g_return_val_if_fail (con != NULL, NULL);
	g_return_val_if_fail (message != NULL, NULL);
	
	if ((reply = dbus_message_new_method_return (message)))
		dbus_message_append_args (reply, DBUS_TYPE_UINT32, &(data->state), DBUS_TYPE_INVALID);

	return reply;
}


/*
 * nm_openswan_dbus_process_helper_config_error
 *
 * Signal the bus that the helper could not get all the configuration information
 * it needed.
 *
 */
static void nm_openswan_dbus_process_helper_config_error (DBusConnection *con, DBusMessage *message, NmOpenSwanData *data)
{
	char *error_item;

	g_return_if_fail (data != NULL);
	g_return_if_fail (con != NULL);
	g_return_if_fail (message != NULL);

	/* Only accept the config info if we're in STARTING state */
	if (data->state != NM_VPN_STATE_STARTING)
		return;

	if (dbus_message_get_args (message, NULL, DBUS_TYPE_STRING, &error_item, DBUS_TYPE_INVALID))
	{
		nm_warning ("openswan helper did not receive adequate configuration information from openswan.  It is missing '%s'.", error_item);
		nm_openswan_dbus_signal_failure (data, NM_DBUS_VPN_SIGNAL_IP_CONFIG_BAD);
	}

	nm_openswan_cancel_helper_timer (data);
	nm_openswan_dbus_handle_stop_vpn (data);
}


/*
 * nm_openswan_dbus_process_helper_ip4_config
 *
 * Signal the bus 
 *
 */
static void nm_openswan_dbus_process_helper_ip4_config (DBusConnection *con, DBusMessage *message, NmOpenSwanData *data)
{
	guint32		ip4_vpn_gateway;
	char *		tundev;
	guint32		ip4_address;
	guint32		ip4_ptp_address;
	guint32		ip4_netmask;
	guint32 *		ip4_dns;
	guint32		ip4_dns_len;
	guint32 *		ip4_nbns;
	guint32		ip4_nbns_len;
	guint32		mss;
	char *		cisco_def_domain;
	char *		cisco_banner;
	gboolean		success = FALSE;
	DBusMessage *	signal;

	g_return_if_fail (data != NULL);
	g_return_if_fail (con != NULL);
	g_return_if_fail (message != NULL);

	/* Only accept the config info if we're in STARTING state */
	if (data->state != NM_VPN_STATE_STARTING)
		return;

	nm_openswan_cancel_helper_timer (data);

	if (!dbus_message_get_args(message, NULL,
	                          DBUS_TYPE_UINT32, &ip4_vpn_gateway,
	                          DBUS_TYPE_STRING, &tundev,
	                          DBUS_TYPE_UINT32, &ip4_address,
	                          DBUS_TYPE_UINT32, &ip4_netmask,
	                          DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &ip4_dns, &ip4_dns_len,
	                          DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &ip4_nbns, &ip4_nbns_len,
	                          DBUS_TYPE_STRING, &cisco_def_domain,
	                          DBUS_TYPE_STRING, &cisco_banner, DBUS_TYPE_INVALID))
		goto out;

	/* For Cisco/openswan, PtP address == local VPN address */
	ip4_ptp_address = ip4_address;

	/* and we don't specify an MSS */
	mss = 0;

	if (!(signal = dbus_message_new_signal (NM_DBUS_PATH_OPENSWAN, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_SIGNAL_IP4_CONFIG)))
		goto out;

	dbus_message_append_args (signal,
	                          DBUS_TYPE_UINT32, &ip4_vpn_gateway,
	                          DBUS_TYPE_STRING, &tundev,
	                          DBUS_TYPE_UINT32, &ip4_address,
	                          DBUS_TYPE_UINT32, &ip4_ptp_address,
	                          DBUS_TYPE_UINT32, &ip4_netmask,
	                          DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &ip4_dns, ip4_dns_len,
	                          DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &ip4_nbns, ip4_nbns_len,
						 DBUS_TYPE_UINT32, &mss,
	                          DBUS_TYPE_STRING, &cisco_def_domain,
	                          DBUS_TYPE_STRING, &cisco_banner, DBUS_TYPE_INVALID);
	if (!dbus_connection_send (data->con, signal, NULL))
	{
		nm_warning ("Could not raise the "NM_DBUS_VPN_SIGNAL_IP4_CONFIG" signal!");
		goto out;
	}

	dbus_message_unref (signal);
	nm_openswan_set_state (data, NM_VPN_STATE_STARTED);
	success = TRUE;

out:
	if (!success)
	{
		nm_warning ("Received invalid IP4 Config information from helper, terminating openswan.");
		nm_openswan_dbus_handle_stop_vpn (data);
	}
}


/*
 * nm_openswan_dbus_message_handler
 *
 * Handle requests for our services.
 *
 */
static DBusHandlerResult nm_openswan_dbus_message_handler (DBusConnection *con, DBusMessage *message, void *user_data)
{
	NmOpenSwanData		*data = (NmOpenSwanData *)user_data;
	const char		*method;
	const char		*path;
	DBusMessage		*reply = NULL;
	gboolean			 handled = TRUE;

	g_return_val_if_fail (data != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
	g_return_val_if_fail (con != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
	g_return_val_if_fail (message != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);

	//get signal from dbus
	method = dbus_message_get_member (message);
	//path for debugging??
	path = dbus_message_get_path (message);
	nm_info ("nm_openswan_dbus_message_handler() got method '%s' for path '%s'.", method, path);

	/* If we aren't ready to accept dbus messages, don't */
	if ((data->state == NM_VPN_STATE_INIT) || (data->state == NM_VPN_STATE_SHUTDOWN))
	{
		nm_warning ("Received dbus messages but couldn't handle them due to INIT or SHUTDOWN states.");
		reply = nm_dbus_create_error_message (message, NM_DBUS_INTERFACE_OPENSWAN, NM_DBUS_VPN_WRONG_STATE,
						"Could not process the request due to current state of STATE_INIT or STATE_SHUTDOWN.");
		goto reply;
	}
	if (strcmp ("startVPN", method) == 0)
		reply = nm_openswan_dbus_start_vpn (con, message, data);
	else if (strcmp ("stopVPN", method) == 0)
		reply = nm_openswan_dbus_stop_vpn (con, message, data);
	else if (strcmp ("getState", method) == 0)
		reply = nm_openswan_dbus_get_state (con, message, data);
	else if (strcmp ("signalConfigError", method) == 0)
		nm_openswan_dbus_process_helper_config_error (con, message, data);
	else if (strcmp ("signalIP4Config", method) == 0)
		nm_openswan_dbus_process_helper_ip4_config (con, message, data);
	else
		handled = FALSE;

reply:
	if (reply)
	{
		dbus_connection_send (con, reply, NULL);
		dbus_message_unref (reply);
	}

	return (handled ? DBUS_HANDLER_RESULT_HANDLED : DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
}


/*
 * nm_openswan_dbus_filter
 *
 * Handle signals from the bus, like NetworkManager network state
 * signals.
 *
 */
static DBusHandlerResult nm_openswan_dbus_filter (DBusConnection *con, DBusMessage *message, void *user_data)
{
	NmOpenSwanData	*data = (NmOpenSwanData *) user_data;
	gboolean		handled = FALSE;
	DBusError		error;

	g_return_val_if_fail (data != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
	g_return_val_if_fail (con != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
	g_return_val_if_fail (message != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);

	if (dbus_message_is_signal (message, DBUS_INTERFACE_DBUS, "NameOwnerChanged"))
	{
		char 	*service;
		char		*old_owner;
		char		*new_owner;

		dbus_error_init (&error);
		if (    dbus_message_get_args (message, &error,
									DBUS_TYPE_STRING, &service,
									DBUS_TYPE_STRING, &old_owner,
									DBUS_TYPE_STRING, &new_owner,
									DBUS_TYPE_INVALID))
		{
			gboolean old_owner_good = (old_owner && (strlen (old_owner) > 0));
			gboolean new_owner_good = (new_owner && (strlen (new_owner) > 0));

			if ((!old_owner_good && new_owner_good) && (strcmp (service, NM_DBUS_SERVICE) == 0))	/* Equivalent to old ServiceCreated signal */
			{
			}
			else if ((old_owner_good && !new_owner_good) && (strcmp (service, NM_DBUS_SERVICE) == 0))	/* Equivalent to old ServiceDeleted signal */
			{
				/* If NM goes away, we don't stick around */
				nm_openswan_dbus_handle_stop_vpn (data);
				g_main_loop_quit (data->loop);
			}
		}
	}
	else if (dbus_message_is_signal (message, NM_DBUS_INTERFACE, "DeviceNoLongerActive"))
	{
		/* If the active device goes down our VPN is certainly not going to work. */
		nm_openswan_dbus_handle_stop_vpn (data);
	}

	return (handled ? DBUS_HANDLER_RESULT_HANDLED : DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
}


/*
 * nm_openswan_dbus_init
 *
 * Grab our connection to the system bus, return NULL if anything goes wrong.
 *
 */
DBusConnection *nm_openswan_dbus_init (NmOpenSwanData *data)
{
	DBusConnection			*connection = NULL;
	DBusError				 error;
	DBusObjectPathVTable	 vtable = { NULL, &nm_openswan_dbus_message_handler, NULL, NULL, NULL, NULL };

	g_return_val_if_fail (data != NULL, NULL);

	dbus_error_init (&error);
	if (!(connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error)))
	{
		nm_warning ("Error connecting to system bus: '%s'", error.message);
		goto out;
	}

	dbus_connection_setup_with_g_main (connection, NULL);

	dbus_error_init (&error);
	dbus_bus_request_name (connection, NM_DBUS_SERVICE_OPENSWAN, 0, &error);
	if (dbus_error_is_set (&error))
	{
		nm_warning ("Could not acquire the dbus service.  dbus_bus_request_name() says: '%s'", error.message);
		goto out;
	}

	if (!dbus_connection_register_object_path (connection, NM_DBUS_PATH_OPENSWAN, &vtable, data))
	{
		nm_warning ("Could not register a dbus handler for nm-openswan-service.  Not enough memory?");
		return NULL;
	}

	if (!dbus_connection_add_filter (connection, nm_openswan_dbus_filter, data, NULL))
		return NULL;

	dbus_error_init (&error);
	//register hooks on the DBUS
	dbus_bus_add_match (connection,
				"type='signal',"
				"interface='" NM_DBUS_INTERFACE "',"
				"sender='" NM_DBUS_SERVICE "',"
				"path='" NM_DBUS_PATH "'",
				&error);
	if (dbus_error_is_set (&error))
		goto out;

	dbus_bus_add_match (connection,
				"type='signal',"
				"interface='" DBUS_INTERFACE_DBUS "',"
				"sender='" DBUS_SERVICE_DBUS "'",
				&error);
	if (dbus_error_is_set (&error))
		goto out;

out:
	if (dbus_error_is_set (&error))
	{
		dbus_error_free (&error);
		connection = NULL;
		nm_warning ("did not get DBUS Connection!");
	}
	else
		nm_info ("Got DBUS Connection");
	return connection;
}



static void sigterm_handler (int signum)
{
	nm_info ("nm-openswan-service caught SIGINT/SIGTERM");
	//g_main_loop_quit (vpn_data->loop);
	exit(EXIT_FAILURE);
}


/*
 * main
 */

int main( int argc, char *argv[] )
{
	struct sigaction	action;
	sigset_t			block_mask;
	NmOpenSwanData     *vpn_data;

	g_type_init ();
	if (!g_thread_supported ())
		g_thread_init (NULL);


	//initialize our data structures
	vpn_data = g_new0 (NmOpenSwanData, 1);
	if (!vpn_data)
	{
		nm_warning("FATAL: Memory Allocation Error!");
		exit(EXIT_FAILURE);
	}
	
	vpn_data->state = NM_VPN_STATE_INIT;
	vpn_data->loop = g_main_loop_new (NULL, FALSE);

	vpn_data->ConnData = g_new0 (VPNConnData, 1);
	vpn_data->ConnData->p1_success = FALSE;
	vpn_data->ConnData->p2_success = FALSE;

	//initilize our connection to dbus
	if (!(vpn_data->con = nm_openswan_dbus_init (vpn_data)))
		exit (EXIT_FAILURE);
	//vpn_data->con = nm_openswan_dbus_init (vpn_data);
	nm_openswan_init_IPC (vpn_data);
	//initialize our socket-based GIOChannel to pluto
	//if (!(nm_openswan_init_IPC (vpn_data)))
		//exit (EXIT_FAILURE);

	//register System Signal handlers (terminate and interrupt)
	action.sa_handler = sigterm_handler;
	sigemptyset (&block_mask);
	action.sa_mask = block_mask;
	action.sa_flags = 0;
	sigaction (SIGINT, &action, NULL);
	sigaction (SIGTERM, &action, NULL);
	
	//check the state of pluto - are there any existing tunnels defined?
	//vpn_data->ConnList_Len = pluto_get_ConnList(vpn_data);
	
	nm_openswan_set_state (vpn_data, NM_VPN_STATE_STOPPED);
	g_main_loop_run (vpn_data->loop);
	
	nm_openswan_dbus_handle_stop_vpn (vpn_data);
	g_main_loop_unref (vpn_data->loop);
	g_free (vpn_data);

	exit (EXIT_SUCCESS);
}
